package dev.fyloz.colorrecipesexplorer.config.initializers

import dev.fyloz.colorrecipesexplorer.config.annotations.RequireDatabase
import dev.fyloz.colorrecipesexplorer.logic.RecipeLogic
import dev.fyloz.colorrecipesexplorer.model.Recipe
import dev.fyloz.colorrecipesexplorer.model.RecipeGroupInformation
import dev.fyloz.colorrecipesexplorer.model.RecipeStep
import dev.fyloz.colorrecipesexplorer.utils.merge
import mu.KotlinLogging
import org.springframework.context.annotation.Configuration

@Configuration
@RequireDatabase
class RecipeInitializer(
    private val recipeLogic: RecipeLogic
) : AbstractInitializer() {
    private val logger = KotlinLogging.logger {}

    override fun initialize() {
        logger.debug("Executing recipe initializer...")
        fixAllPositions()
    }

    private fun fixAllPositions() {
        logger.debug("Validating recipes steps positions...")

        recipeLogic.getAll()
            .forEach(this::fixRecipePositions)

        logger.debug("Recipes steps positions are valid!")
    }

    private fun fixRecipePositions(recipe: Recipe) {
        val fixedGroupInformation = recipe.groupsInformation
            .filter { it.steps != null }
            .filter { groupInfo -> groupInfo.steps!!.any { it.position == 0 } }
            .map { fixGroupInformationPositions(recipe, it) }

        val updatedGroupInformation = recipe.groupsInformation.merge(fixedGroupInformation) { it.id }

        with(recipe.copy(groupsInformation = updatedGroupInformation.toMutableSet())) {
            recipeLogic.update(this)
        }
    }

    private fun fixGroupInformationPositions(
        recipe: Recipe,
        groupInformation: RecipeGroupInformation
    ): RecipeGroupInformation {
        val steps = groupInformation.steps!!
        val maxPosition = steps.maxOf { it.position }

        logger.warn("Recipe ${recipe.id} (${recipe.name}) has invalid positions:")

        val invalidRecipeSteps = steps.filter { it.position == 0 }
        val fixedRecipeSteps = increaseRecipeStepsPosition(groupInformation, invalidRecipeSteps, maxPosition + 1)
        val updatedRecipeSteps = steps.merge(fixedRecipeSteps) { it.id }

        return groupInformation.copy(steps = updatedRecipeSteps.toMutableSet())
    }

    private fun increaseRecipeStepsPosition(
        groupInformation: RecipeGroupInformation,
        recipeSteps: Iterable<RecipeStep>,
        firstPosition: Int
    ) =
        recipeSteps
            .mapIndexed { index, recipeStep -> recipeStep.copy(position = firstPosition + index) }
            .onEach {
                logger.info("\tPosition of step ${it.id} (group: ${groupInformation.group.name}) has been set to ${it.position}")
            }
}