package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.EntityDto
import dev.fyloz.colorrecipesexplorer.model.ModelEntity
import dev.fyloz.colorrecipesexplorer.model.NamedModelEntity
import dev.fyloz.colorrecipesexplorer.repository.NamedJpaRepository
import io.jsonwebtoken.lang.Assert
import org.springframework.data.jpa.repository.JpaRepository
import org.springframework.data.repository.findByIdOrNull

/**
 * A service implementing the basics CRUD operations for the given entities.
 *
 * @param E The entity type
 * @param R The entity repository type
 */
interface OldService<E, R : JpaRepository<E, *>> {
    val repository: R

    /** Gets all entities. */
    fun getAll(): Collection<E>

    /** Saves a given [entity]. */
    fun save(entity: E): E

    /** Updates a given [entity]. */
    fun update(entity: E): E

    /** Deletes a given [entity]. */
    fun delete(entity: E)
}

/** A service for entities implementing the [ModelEntity] interface. This service add supports for numeric identifiers. */
interface ModelService<E : ModelEntity, R : JpaRepository<E, *>> : OldService<E, R> {
    /** Checks if an entity with the given [id] exists. */
    fun existsById(id: Long): Boolean

    /** Gets the entity with the given [id]. */
    fun getById(id: Long): E

    /** Deletes the entity with the given [id]. */
    fun deleteById(id: Long)
}

/** A service for entities implementing the [NamedModelEntity] interface. This service add supports for name identifiers. */
interface NamedModelService<E : NamedModelEntity, R : JpaRepository<E, *>> : ModelService<E, R> {
    /** Checks if an entity with the given [name] exists. */
    fun existsByName(name: String): Boolean

    /** Gets the entity with the given [name]. */
    fun getByName(name: String): E
}


abstract class AbstractService<E, R : JpaRepository<E, *>>(override val repository: R) : OldService<E, R> {
    override fun getAll(): Collection<E> = repository.findAll()
    override fun save(entity: E): E = repository.save(entity)
    override fun update(entity: E): E = repository.save(entity)
    override fun delete(entity: E) = repository.delete(entity)
}

abstract class AbstractModelService<E : ModelEntity, R : JpaRepository<E, Long>>(repository: R) :
    AbstractService<E, R>(repository), ModelService<E, R> {
    protected abstract fun idNotFoundException(id: Long): NotFoundException
    protected abstract fun idAlreadyExistsException(id: Long): AlreadyExistsException

    override fun existsById(id: Long): Boolean = repository.existsById(id)
    override fun getById(id: Long): E = repository.findByIdOrNull(id) ?: throw idNotFoundException(id)

    override fun save(entity: E): E {
        if (entity.id != null && existsById(entity.id!!))
            throw idAlreadyExistsException(entity.id!!)
        return super.save(entity)
    }

    override fun update(entity: E): E {
        assertId(entity.id)
        if (!existsById(entity.id!!))
            throw idNotFoundException(entity.id!!)
        return super.update(entity)
    }

    override fun deleteById(id: Long) =
        delete(getById(id)) // Use delete(entity) to prevent code duplication and to ease testing

    protected fun assertId(id: Long?) {
        Assert.notNull(id, "${javaClass.simpleName}.update() was called with a null identifier")
    }
}

abstract class AbstractNamedModelService<E : NamedModelEntity, R : NamedJpaRepository<E>>(repository: R) :
    AbstractModelService<E, R>(repository), NamedModelService<E, R> {
    protected abstract fun nameNotFoundException(name: String): NotFoundException
    protected abstract fun nameAlreadyExistsException(name: String): AlreadyExistsException

    override fun existsByName(name: String): Boolean = repository.existsByName(name)
    override fun getByName(name: String): E = repository.findByName(name) ?: throw nameNotFoundException(name)

    override fun save(entity: E): E {
        if (existsByName(entity.name))
            throw nameAlreadyExistsException(entity.name)
        return super.save(entity)
    }

    override fun update(entity: E): E {
        assertId(entity.id)
        assertName(entity.name)
        with(repository.findByName(entity.name)) {
            if (this != null && id != entity.id)
                throw nameAlreadyExistsException(entity.name)
        }
        return super.update(entity)
    }

    private fun assertName(name: String) {
        Assert.notNull(name, "${javaClass.simpleName}.update() was called with a null name")
    }
}

/**
 * A service that will receive *external* interactions, from the REST API, for example.
 *
 * @param E The entity type
 * @param S The entity save DTO type
 * @param U The entity update DTO type
 */
interface ExternalService<E, S : EntityDto<E>, U : EntityDto<E>, O, R : JpaRepository<E, *>> : OldService<E, R> {
    /** Gets all entities mapped to their output model. */
    fun getAllForOutput(): Collection<O>

    /** Saves a given [entity]. */
    fun save(entity: S): E = save(entity.toEntity())

    /** Updates a given [entity]. */
    fun update(entity: U): E

    /** Convert the given entity to its output model. */
    fun E.toOutput(): O
}

/** An [ExternalService] for entities implementing the [ModelEntity] interface. */
interface ExternalModelService<E : ModelEntity, S : EntityDto<E>, U : EntityDto<E>, O, R : JpaRepository<E, *>> :
    ModelService<E, R>, ExternalService<E, S, U, O, R> {
    /** Gets the entity with the given [id] mapped to its output model. */
    fun getByIdForOutput(id: Long): O
}

/** An [ExternalService] for entities implementing the [NamedModelEntity] interface. */
interface ExternalNamedModelService<E : NamedModelEntity, S : EntityDto<E>, U : EntityDto<E>, O, R : JpaRepository<E, *>> :
    NamedModelService<E, R>, ExternalModelService<E, S, U, O, R>

/** An [AbstractService] with the functionalities of a [ExternalService]. */
@Suppress("unused")
abstract class AbstractExternalService<E, S : EntityDto<E>, U : EntityDto<E>, O, R : JpaRepository<E, *>>(repository: R) :
    AbstractService<E, R>(repository), ExternalService<E, S, U, O, R> {
    override fun getAllForOutput() =
        getAll().map { it.toOutput() }
}

/** An [AbstractModelService] with the functionalities of a [ExternalService]. */
abstract class AbstractExternalModelService<E : ModelEntity, S : EntityDto<E>, U : EntityDto<E>, O, R : JpaRepository<E, Long>>(
    repository: R
) : AbstractModelService<E, R>(repository), ExternalModelService<E, S, U, O, R> {
    override fun getAllForOutput() =
        getAll().map { it.toOutput() }

    override fun getByIdForOutput(id: Long) =
        getById(id).toOutput()
}

/** An [AbstractNamedModelService] with the functionalities of a [ExternalService]. */
abstract class AbstractExternalNamedModelService<E : NamedModelEntity, S : EntityDto<E>, U : EntityDto<E>, O, R : NamedJpaRepository<E>>(
    repository: R
) : AbstractNamedModelService<E, R>(repository), ExternalNamedModelService<E, S, U, O, R> {
    override fun getAllForOutput() =
        getAll().map { it.toOutput() }

    override fun getByIdForOutput(id: Long) =
        getById(id).toOutput()
}
