package dev.fyloz.colorrecipesexplorer.rest

import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeEditRecipes
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeViewRecipes
import dev.fyloz.colorrecipesexplorer.logic.MixLogic
import dev.fyloz.colorrecipesexplorer.logic.RecipeImageLogic
import dev.fyloz.colorrecipesexplorer.logic.RecipeLogic
import dev.fyloz.colorrecipesexplorer.model.*
import org.springframework.context.annotation.Profile
import org.springframework.http.MediaType
import org.springframework.http.ResponseEntity
import org.springframework.security.access.prepost.PreAuthorize
import org.springframework.web.bind.annotation.*
import org.springframework.web.multipart.MultipartFile
import javax.validation.Valid


private const val RECIPE_CONTROLLER_PATH = "api/recipe"

@RestController
@RequestMapping(RECIPE_CONTROLLER_PATH)
@Profile("!emergency")
@PreAuthorizeViewRecipes
class RecipeController(
    private val recipeLogic: RecipeLogic,
    private val recipeImageLogic: RecipeImageLogic
) {
    @GetMapping
    fun getAll(@RequestParam(required = false) name: String?) =
        if (name == null)
            ok(recipeLogic.getAllForOutput())
        else
            ok(with(recipeLogic) {
                getAllByName(name).map { it.toOutput() }
            })

    @GetMapping("{id}")
    fun getById(@PathVariable id: Long) =
        ok(recipeLogic.getByIdForOutput(id))

    @PostMapping
    @PreAuthorizeEditRecipes
    fun save(@Valid @RequestBody recipe: RecipeSaveDto) =
        created<RecipeOutputDto>(RECIPE_CONTROLLER_PATH) {
            with(recipeLogic) {
                save(recipe).toOutput()
            }
        }

    @PutMapping
    @PreAuthorizeEditRecipes
    fun update(@Valid @RequestBody recipe: RecipeUpdateDto) =
        noContent {
            recipeLogic.update(recipe)
        }

    @PutMapping("public")
    @PreAuthorize("hasAuthority('EDIT_RECIPES_PUBLIC_DATA')")
    fun updatePublicData(@Valid @RequestBody publicDataDto: RecipePublicDataDto) =
        noContent {
            recipeLogic.updatePublicData(publicDataDto)
        }

    @DeleteMapping("{id}")
    @PreAuthorizeEditRecipes
    fun deleteById(@PathVariable id: Long) =
        noContent {
            recipeLogic.deleteById(id)
        }

    @PutMapping("{recipeId}/image", consumes = [MediaType.MULTIPART_FORM_DATA_VALUE])
    @PreAuthorizeEditRecipes
    fun downloadImage(@PathVariable recipeId: Long, image: MultipartFile): ResponseEntity<RecipeOutputDto> {
        recipeImageLogic.download(image, recipeLogic.getById(recipeId))
        return getById(recipeId)
    }

    @DeleteMapping("{recipeId}/image/{name}")
    @PreAuthorizeEditRecipes
    fun deleteImage(@PathVariable recipeId: Long, @PathVariable name: String) =
        noContent {
            recipeImageLogic.delete(recipeLogic.getById(recipeId), name)
        }
}
