package dev.fyloz.colorrecipesexplorer.utils

import dev.fyloz.colorrecipesexplorer.dtos.EntityDto

/** Returns a list containing the result of the given [transform] applied to each item of the [Iterable]. If the given [transform] throws, the [Throwable] will be passed to the given [throwableConsumer]. */
inline fun <T, R, reified E : Throwable> Iterable<T>.mapMayThrow(
    throwableConsumer: (E) -> Unit = {},
    transform: (T) -> R
): List<R> = this.mapNotNull {
    try {
        transform(it)
    } catch (th: Throwable) {
        if (th is E) {
            throwableConsumer(th)
            null
        } else {
            throw th
        }
    }
}

/** Find duplicated elements in the given [Iterable]. */
fun <T> Iterable<T>.findDuplicated() =
    this.groupBy { it }
        .filter { it.value.count() > 1 }
        .map { it.key }

/** Check if the given [Int] [Iterable] has gaps between each element. */
fun Iterable<Int>.hasGaps() =
    this.sorted()
        .filterIndexed { index, it -> it != index + 1 }
        .isNotEmpty()

/** Clears and fills the given [MutableCollection] with the given [elements]. */
fun <T> MutableCollection<T>.setAll(elements: Collection<T>) {
    this.clear()
    this.addAll(elements)
}

/** Removes and returns all elements of a [MutableCollection] matching the given [predicate]. */
inline fun <T> MutableCollection<T>.excludeAll(predicate: (T) -> Boolean): Iterable<T> {
    val matching = this.filter(predicate)
    this.removeAll(matching.toSet())
    return matching
}

/**
 * Merge two [EntityDto] [Iterable]s and prevent duplication of their ids.
 * In case of collision, the items from the [other] iterable will be taken.
 */
@JvmName("mergeDto")
fun <T : EntityDto> Iterable<T>.merge(other: Iterable<T>) =
    this.merge(other) { it.id }

/**
 * Merge two [Iterable]s and prevent duplication of the keys determined by the given [keyMapper].
 * In case of collision, the items from the [other] iterable will be taken.
 */
fun <T, K> Iterable<T>.merge(other: Iterable<T>, keyMapper: (T) -> K) =
    this.associateBy { keyMapper(it) }
        .filter { pair -> other.all { keyMapper(it) != pair.key } }
        .map { it.value }
        .plus(other)
