package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.dtos.*
import dev.fyloz.colorrecipesexplorer.model.Recipe
import dev.fyloz.colorrecipesexplorer.model.company
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import kotlin.test.assertEquals

class DefaultInventoryLogicTest {
    private val materialLogicMock = mockk<MaterialLogic>()
    private val mixLogicMock = mockk<MixLogic>()

    private val inventoryLogic = spyk(DefaultInventoryLogic(materialLogicMock, mixLogicMock))

    private val defaultInventoryQuantity = 5000f
    private val materialType = MaterialTypeDto(1L, "Unit test material type", "UTMT", false)
    private val material = MaterialDto(1L, "Unit test material", defaultInventoryQuantity, false, materialType)

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    @Test
    fun add_collection_normalBehavior_callsAddSingleForEachQuantity() {
        // Arrange
        val quantities = setOf(
            MaterialQuantityDto(1L, 1000f),
            MaterialQuantityDto(2L, 2000f)
        )
        val expectedQuantities = quantities.map { it.copy(quantity = it.quantity + defaultInventoryQuantity) }

        every { inventoryLogic.add(any<MaterialQuantityDto>()) } answers { expectedQuantities[this.nArgs].quantity }

        // Act
        inventoryLogic.add(quantities)

        // Assert
        verify {
            quantities.forEach {
                inventoryLogic.add(it)
            }
        }
    }

    @Test
    fun add_collection_normalBehavior_returnsFromAddSingle() {
        // Arrange
        val quantities = setOf(
            MaterialQuantityDto(1L, 1000f),
            MaterialQuantityDto(2L, 2000f)
        )
        val expectedQuantities = quantities.map { it.copy(quantity = it.quantity + defaultInventoryQuantity) }

        every { inventoryLogic.add(any<MaterialQuantityDto>()) } returnsMany (expectedQuantities.map { it.quantity })

        // Act
        val actualQuantities = inventoryLogic.add(quantities)

        // Assert
        assertEquals(expectedQuantities, actualQuantities)
    }

    @Test
    fun add_single_normalBehavior_callsUpdateQuantityInMaterialLogic() {
        // Arrange
        every { materialLogicMock.getById(any()) } returns material
        every { materialLogicMock.updateQuantity(any(), any()) } answers {
            this.secondArg<Float>() + defaultInventoryQuantity
        }

        val quantity = MaterialQuantityDto(1L, 1000f)

        // Act
        inventoryLogic.add(quantity)

        // Assert
        verify {
            materialLogicMock.updateQuantity(material, quantity.quantity)
        }
    }

    @Test
    fun deductMix_normalBehavior_callsDeductWithMixMaterials() {
        // Arrange
        val company = CompanyDto(1L, "Unit test company")
        val recipe = Recipe(
            1L,
            "Unit test recipe",
            "Unit test recipe",
            "FFFFFF",
            0xf,
            null,
            null,
            "Remark",
            company(company),
            mutableListOf(),
            setOf()
        )
        val mixType = MixTypeDto(1L, "Unit test mix type", material)
        val mixMaterial = MixMaterialDto(1L, material, 1000f, 1)
        val mix = MixDto(1L, null, recipe, mixType, setOf(mixMaterial))

        val dto = MixDeductDto(mix.id, 2f)
        val expectedQuantities = listOf(MaterialQuantityDto(material.id, mixMaterial.quantity * dto.ratio))

        every { mixLogicMock.getById(any()) } returns mix
        every { inventoryLogic.deduct(any<Collection<MaterialQuantityDto>>()) } returns listOf()

        // Act
        inventoryLogic.deductMix(dto)

        // Assert
        verify {
            inventoryLogic.deduct(expectedQuantities)
        }
    }

    @Test
    fun deductMix_normalBehavior_returnsFromDeduct() {
        // Arrange
        val company = CompanyDto(1L, "Unit test company")
        val recipe = Recipe(
            1L,
            "Unit test recipe",
            "Unit test recipe",
            "FFFFFF",
            0xf,
            null,
            null,
            "Remark",
            company(company),
            mutableListOf(),
            setOf()
        )
        val mixType = MixTypeDto(1L, "Unit test mix type", material)
        val mixMaterial = MixMaterialDto(1L, material, 1000f, 1)
        val mix = MixDto(1L, null, recipe, mixType, setOf(mixMaterial))

        val dto = MixDeductDto(mix.id, 2f)
        val expectedQuantities = listOf(MaterialQuantityDto(material.id, mixMaterial.quantity * dto.ratio))

        every { mixLogicMock.getById(any()) } returns mix
        every { inventoryLogic.deduct(any<Collection<MaterialQuantityDto>>()) } returns expectedQuantities

        // Act
        val actualQuantities = inventoryLogic.deductMix(dto)

        // Assert
        assertEquals(expectedQuantities, actualQuantities)
    }

    @Test
    fun deduct_collection_normalBehavior_callsDeductForEachQuantity() {
        // Arrange
        val quantities = setOf(
            MaterialQuantityDto(1L, 1000f),
            MaterialQuantityDto(2L, 2000f)
        )

        every { inventoryLogic.deduct(any<MaterialQuantityDto>()) } answers {
            defaultInventoryQuantity - firstArg<MaterialQuantityDto>().quantity
        }

        // Act
        inventoryLogic.deduct(quantities)

        // Assert
        verify {
            quantities.forEach {
                inventoryLogic.deduct(it)
            }
        }
    }

    @Test
    fun deduct_collection_normalBehavior_returnsFromDeduct() {
        // Arrange
        val quantities = setOf(
            MaterialQuantityDto(1L, 1000f),
            MaterialQuantityDto(2L, 2000f)
        )

        every { inventoryLogic.deduct(any<MaterialQuantityDto>()) } answers {
            defaultInventoryQuantity - firstArg<MaterialQuantityDto>().quantity
        }

        val expectedQuantities =
            quantities.map { MaterialQuantityDto(it.materialId, defaultInventoryQuantity - it.quantity) }

        // Act
        val actualQuantities = inventoryLogic.deduct(quantities)

        // Assert
        assertEquals(expectedQuantities, actualQuantities)
    }

    @Test
    fun deduct_collection_notEnoughInventory_throwsMultiplesNotEnoughInventoryException() {
        // Arrange
        val quantities = setOf(
            MaterialQuantityDto(1L, 1000f),
            MaterialQuantityDto(2L, 2000f)
        )

        every { inventoryLogic.deduct(any<MaterialQuantityDto>()) } throws NotEnoughInventoryException(1000f, material)

        // Act
        // Assert
        assertThrows<MultiplesNotEnoughInventoryException> { inventoryLogic.deduct(quantities) }
    }

    @Test
    fun deduct_single_normalBehavior_callsUpdateQuantityInMaterialLogic() {
        // Arrange
        val quantity = MaterialQuantityDto(material.id, 1000f)

        every { materialLogicMock.getById(any()) } returns material
        every { materialLogicMock.updateQuantity(any(), any()) } returns defaultInventoryQuantity - quantity.quantity

        // Act
        inventoryLogic.deduct(quantity)

        // Assert
        verify {
            materialLogicMock.getById(quantity.materialId)
            materialLogicMock.updateQuantity(material, -quantity.quantity)
        }
        confirmVerified(materialLogicMock)
    }

    @Test
    fun deduct_single_notEnoughInventory_throwsNotEnoughInventoryException() {
        // Arrange
        val quantity = MaterialQuantityDto(material.id, material.inventoryQuantity + 1000f)

        every { materialLogicMock.getById(any()) } returns material
        every { materialLogicMock.updateQuantity(any(), any()) } returns defaultInventoryQuantity - quantity.quantity

        // Act
        // Assert
        assertThrows<NotEnoughInventoryException> { inventoryLogic.deduct(quantity) }
    }
}