package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.dtos.MaterialDto
import dev.fyloz.colorrecipesexplorer.dtos.MaterialTypeDto
import dev.fyloz.colorrecipesexplorer.dtos.MixTypeDto
import dev.fyloz.colorrecipesexplorer.exception.CannotDeleteException
import dev.fyloz.colorrecipesexplorer.service.MixTypeService
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import kotlin.math.exp
import kotlin.test.assertEquals

class DefaultMixTypeLogicTest {
    private val mixTypeServiceMock = mockk<MixTypeService>()
    private val materialLogicMock = mockk<MaterialLogic>()

    private val mixTypeLogic = spyk(DefaultMixTypeLogic(mixTypeServiceMock, materialLogicMock))

    private val materialType = MaterialTypeDto(1L, "Unit test material type", "UTMT", false)
    private val material = MaterialDto(1L, "Unit test material", 1000f, true, materialType)
    private val mixType = MixTypeDto(id = 1L, name = "Unit test mix type", material)

    @AfterEach
    fun afterEach() {
        clearAllMocks()
    }

    @Test
    fun getOrCreateForNameAndMaterialType_normalBehavior_returnsFromService() {
        // Arrange
        every { mixTypeServiceMock.getByNameAndMaterialType(any(), any()) } returns mixType

        // Act
        val actualMixType = mixTypeLogic.getOrCreateForNameAndMaterialType(mixType.name, materialType)

        // Assert
        assertEquals(mixType, actualMixType)
    }

    @Test
    fun getOrCreateForNameAndMaterialType_notFound_callsSave() {
        // Arrange
        every { mixTypeServiceMock.getByNameAndMaterialType(any(), any()) } returns null
        every { materialLogicMock.save(any<MaterialDto>()) } returns material
        every { mixTypeLogic.save(any()) } returnsArgument 0

        val expectedMixType = mixType.copy(id = 0L)

        // Act
        mixTypeLogic.getOrCreateForNameAndMaterialType(mixType.name, materialType)

        // Assert
        verify {
            mixTypeLogic.save(expectedMixType)
        }
    }

    @Test
    fun getOrCreateForNameAndMaterialType_notFound_returnsFromSave() {
        // Arrange
        every { mixTypeServiceMock.getByNameAndMaterialType(any(), any()) } returns null
        every { materialLogicMock.save(any<MaterialDto>()) } returns material
        every { mixTypeLogic.save(any()) } returnsArgument 0

        val expectedMixType = mixType.copy(id = 0L)

        // Act
        val actualMixType = mixTypeLogic.getOrCreateForNameAndMaterialType(mixType.name, materialType)

        // Assert
        assertEquals(expectedMixType, actualMixType)
    }

    @Test
    fun updateOrCreateForNameAndMaterialType_mixTypeShared_callsSave() {
        // Arrange
        every { mixTypeServiceMock.isShared(any()) } returns true
        every { materialLogicMock.save(any<MaterialDto>()) } returns material
        every { mixTypeLogic.save(any()) } returnsArgument 0

        val expectedMixType = mixType.copy(id = 0L, name = "${mixType.name} updated")

        // Act
        mixTypeLogic.updateOrCreateForNameAndMaterialType(mixType, expectedMixType.name, materialType)

        // Assert
        verify {
            mixTypeLogic.save(expectedMixType)
        }
    }

    @Test
    fun updateOrCreateForNameAndMaterialType_mixTypeShared_returnsFromSave() {
        // Arrange
        every { mixTypeServiceMock.isShared(any()) } returns true
        every { materialLogicMock.save(any<MaterialDto>()) } returns material
        every { mixTypeLogic.save(any()) } returnsArgument 0

        val expectedMixType = mixType.copy(id = 0L, name = "${mixType.name} updated")

        // Act
        val actualMixType = mixTypeLogic.updateOrCreateForNameAndMaterialType(mixType, expectedMixType.name, materialType)

        // Assert
        assertEquals(expectedMixType, actualMixType)
    }

    @Test
    fun updateOrCreateForNameAndMaterialType_mixTypeNotShared_callsUpdate() {
        // Arrange
        every { mixTypeServiceMock.isShared(any()) } returns false
        every { materialLogicMock.update(any<MaterialDto>()) } returns material
        every { mixTypeLogic.update(any()) } returnsArgument 0

        val expectedMixType = mixType.copy(name = "${mixType.name} updated")

        // Act
        mixTypeLogic.updateOrCreateForNameAndMaterialType(mixType, expectedMixType.name, materialType)

        // Assert
        verify {
            mixTypeLogic.update(expectedMixType)
        }
    }

    @Test
    fun updateOrCreateForNameAndMaterialType_mixTypeNotShared_returnsFromUpdate() {
        // Arrange
        every { mixTypeServiceMock.isShared(any()) } returns false
        every { materialLogicMock.update(any<MaterialDto>()) } returns material
        every { mixTypeLogic.update(any()) } returnsArgument 0

        val expectedMixType = mixType.copy(name = "${mixType.name} updated")

        // Act
        val actualMixType = mixTypeLogic.updateOrCreateForNameAndMaterialType(mixType, expectedMixType.name, materialType)

        // Assert
        assertEquals(expectedMixType, actualMixType)
    }

    @Test
    fun deleteById_usedByMix_throwsCannotDeleteException() {
        // Arrange
        every { mixTypeServiceMock.isUsedByMixes(any()) } returns true

        // Act
        // Assert
        assertThrows<CannotDeleteException> { mixTypeLogic.deleteById(mixType.id) }
    }
}