package dev.fyloz.colorrecipesexplorer.logic

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.dtos.MixLocationDto
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.logic.config.ConfigurationLogic
import dev.fyloz.colorrecipesexplorer.logic.files.CachedFile
import dev.fyloz.colorrecipesexplorer.logic.files.WriteableFileLogic
import dev.fyloz.colorrecipesexplorer.logic.users.GroupLogic
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.model.account.group
import dev.fyloz.colorrecipesexplorer.repository.RecipeRepository
import dev.fyloz.colorrecipesexplorer.utils.FilePath
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.TestInstance
import org.junit.jupiter.api.assertThrows
import org.springframework.mock.web.MockMultipartFile
import org.springframework.web.multipart.MultipartFile
import java.time.LocalDate
import java.time.Period
import kotlin.test.*

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class RecipeLogicTest :
    AbstractExternalModelServiceTest<Recipe, RecipeSaveDto, RecipeUpdateDto, RecipeLogic, RecipeRepository>() {
    override val repository: RecipeRepository = mock()
    private val companyLogic: CompanyLogic = mock()
    private val mixService: MixLogic = mock()
    private val groupService: GroupLogic = mock()
    private val recipeStepService: RecipeStepLogic = mock()
    private val configService: ConfigurationLogic = mock()
    override val logic: RecipeLogic =
        spy(
            DefaultRecipeLogic(
                repository,
                companyLogic,
                mixService,
                recipeStepService,
                groupService,
                mock(),
                configService
            )
        )

    private val company: Company = company(id = 0L)
    override val entity: Recipe = recipe(id = 0L, name = "recipe", company = company)
    override val anotherEntity: Recipe = recipe(id = 1L, name = "another recipe", company = company)
    override val entitySaveDto: RecipeSaveDto = spy(recipeSaveDto(name = entity.name, companyId = entity.company.id!!))
    override val entityUpdateDto: RecipeUpdateDto = spy(recipeUpdateDto(id = entity.id!!, name = entity.name))

    @AfterEach
    override fun afterEach() {
        reset(companyLogic, mixService)
        super.afterEach()
    }

    // existsByCompany()

    @Test
    fun `existsByCompany() returns true when at least one recipe exists for the given company`() {
        whenever(repository.existsByCompany(company)).doReturn(true)

        val found = logic.existsByCompany(company)

        assertTrue(found)
    }

    @Test
    fun `existsByCompany() returns false when no recipe exists for the given company`() {
        whenever(repository.existsByCompany(company)).doReturn(false)

        val found = logic.existsByCompany(company)

        assertFalse(found)
    }

    // existsByNameAndCompany()

    @Test
    fun `existsByNameAndCompany() returns if a recipe exists for the given name and company in the repository`() {
        setOf(true, false).forEach {
            whenever(repository.existsByNameAndCompany(entity.name, company)).doReturn(it)

            val exists = logic.existsByNameAndCompany(entity.name, company)

            assertEquals(it, exists)
        }
    }

    // isApprobationExpired()

    @Test
    fun `isApprobationExpired() returns false when the approbation date of the given recipe is within the configured period`() {
        val period = Period.ofMonths(4)
        val recipe = recipe(approbationDate = LocalDate.now())

        whenever(configService.getContent(ConfigurationType.RECIPE_APPROBATION_EXPIRATION)).doReturn(period.toString())

        val approbationExpired = logic.isApprobationExpired(recipe)

        assertNotNull(approbationExpired)
        assertFalse(approbationExpired)
    }

    @Test
    fun `isApprobationExpired() returns true when the approbation date of the given recipe is outside the configured period`() {
        val period = Period.ofMonths(4)
        val recipe = recipe(approbationDate = LocalDate.now().minus(period).minusMonths(1))

        whenever(configService.getContent(ConfigurationType.RECIPE_APPROBATION_EXPIRATION)).doReturn(period.toString())

        val approbationExpired = logic.isApprobationExpired(recipe)

        assertNotNull(approbationExpired)
        assertTrue(approbationExpired)
    }

    @Test
    fun `isApprobationExpired() returns null when the given recipe as no approbation date`() {
        val period = Period.ofMonths(4)
        val recipe = recipe(approbationDate = null)

        whenever(configService.getContent(ConfigurationType.RECIPE_APPROBATION_EXPIRATION)).doReturn(period.toString())

        val approbationExpired = logic.isApprobationExpired(recipe)

        assertNull(approbationExpired)
    }

    // getAllByName()

    @Test
    fun `getAllByName() returns the recipes with the given name`() {
        val recipes = listOf(entity, anotherEntity)

        whenever(repository.findAllByName(entity.name)).doReturn(recipes)

        val found = logic.getAllByName(entity.name)

        assertEquals(recipes, found)
    }

    // getAllByCompany()

    @Test
    fun `getAllByCompany() returns the recipes with the given company`() {
        val companies = listOf(entity, anotherEntity)
        whenever(repository.findAllByCompany(company)).doReturn(companies)

        val found = logic.getAllByCompany(company)

        assertEquals(companies, found)
    }

    // save()

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        whenever(companyLogic.getById(company.id!!)).doReturn(companyDto(company))
        doReturn(false).whenever(logic).existsByNameAndCompany(entity.name, company)
        withBaseSaveDtoTest(entity, entitySaveDto, logic, { argThat { this.id == null && this.color == color } })
    }

    @Test
    fun `save(dto) throw AlreadyExistsException when a recipe with the given name and company exists in the repository`() {
        whenever(companyLogic.getById(company.id!!)).doReturn(companyDto(company))
        doReturn(true).whenever(logic).existsByNameAndCompany(entity.name, company)

        with(assertThrows<AlreadyExistsException> { logic.save(entitySaveDto) }) {
            this.assertErrorCode("company-name")
        }
    }

    // update()

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() {
        doReturn(false).whenever(logic).existsByNameAndCompany(entity.name, company)
        withBaseUpdateDtoTest(entity, entityUpdateDto, logic, { any() })
    }

    @Test
    fun `update(dto) throws AlreadyExistsException when a recipe exists for the given name and company`() {
        val name = "another recipe"

        doReturn(entity).whenever(logic).getById(entity.id!!)
        doReturn(true).whenever(logic).existsByNameAndCompany(name, company)
        doReturn(name).whenever(entityUpdateDto).name

        with(assertThrows<AlreadyExistsException> { logic.update(entityUpdateDto) }) {
            this.assertErrorCode("company-name")
        }
    }

    // updatePublicData()

    @Test
    fun `updatePublicData() updates the notes of a recipe groups information according to the RecipePublicDataDto`() {
        val recipe = recipe(
            id = 0L, groupsInformation = setOf(
                recipeGroupInformation(id = 0L, group = group(id = 1L), note = "Old note"),
                recipeGroupInformation(id = 1L, group = group(id = 2L), note = "Another note"),
                recipeGroupInformation(id = 2L, group = group(id = 3L), note = "Up to date note")
            )
        )
        val notes = setOf(
            noteDto(groupId = 1, content = "Note 1"),
            noteDto(groupId = 2, content = null)
        )
        val publicData = recipePublicDataDto(recipeId = recipe.id!!, notes = notes)

        doReturn(recipe).whenever(logic).getById(recipe.id!!)
        doAnswer { it.arguments[0] }.whenever(logic).update(any<Recipe>())

        logic.updatePublicData(publicData)

        verify(logic).update(argThat<Recipe> {
            assertTrue { this.groupsInformation.first { it.group.id == 1L }.note == notes.first { it.groupId == 1L }.content }
            assertTrue { this.groupsInformation.first { it.group.id == 2L }.note == notes.first { it.groupId == 2L }.content }
            assertTrue { this.groupsInformation.any { it.group.id == 3L } && this.groupsInformation.first { it.group.id == 3L }.note == null }
            true
        })
        verify(mixService, times(0)).updateLocations(any())
    }

    @Test
    fun `updatePublicData() update the location of a recipe mixes in the mix logic according to the RecipePublicDataDto`() {
        val publicData = recipePublicDataDto(
            mixesLocation = setOf(
                MixLocationDto(mixId = 0L, location = "Loc 1"),
                MixLocationDto(mixId = 1L, location = "Loc 2")
            )
        )

        logic.updatePublicData(publicData)

        verify(mixService).updateLocations(publicData.mixesLocation!!)
        verify(logic, times(0)).update(any<Recipe>())
    }

    // addMix()

    @Test
    fun `addMix() adds the given mix to the given recipe and updates it`() {
        val mix = mix(id = 0L)
        val recipe = recipe(id = 0L, mixes = mutableListOf())

        doAnswer { it.arguments[0] }.whenever(logic).update(any<Recipe>())

        val found = logic.addMix(recipe, mix)

        verify(logic).update(any<Recipe>())

        assertEquals(recipe.id, found.id)
        assertTrue(found.mixes.contains(mix))
    }

    // removeMix()

    @Test
    fun `removeMix() removes the given mix from its recipe and updates it`() {
        val recipe = recipe(id = 0L, mixes = mutableListOf())
        val mix = mix(id = 0L, recipe = recipe)
        recipe.mixes.add(mix)

        doAnswer { it.arguments[0] }.whenever(logic).update(any<Recipe>())

        val found = logic.removeMix(mix)

        verify(logic).update(any<Recipe>())

        assertEquals(recipe.id, found.id)
        assertFalse(found.mixes.contains(mix))
    }
}

private class RecipeImageServiceTestContext {
    val fileService = mockk<WriteableFileLogic> {
        every { write(any<MultipartFile>(), any(), any()) } just Runs
        every { delete(any()) } just Runs
    }
    val recipeImageService = spyk(DefaultRecipeImageLogic(fileService))
    val recipe = spyk(recipe())
    val recipeImagesIds = setOf(1L, 10L, 21L)
    val recipeImagesNames = recipeImagesIds.map { it.imageName }.toSet()
    val recipeImagesFiles = recipeImagesNames.map { CachedFile(it, FilePath(it), true) }

    val Long.imageName
        get() = "${recipe.name}$RECIPE_IMAGE_ID_DELIMITER$this"

    val String.imagePath
        get() = "${recipe.imagesDirectoryPath}/$this$RECIPE_IMAGE_EXTENSION"
}

class RecipeImageLogicTest {
    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    private fun test(test: RecipeImageServiceTestContext.() -> Unit) {
        RecipeImageServiceTestContext().test()
    }

    // getAllImages()

    @Test
    fun `getAllImages() returns a Set containing the name of every files in the recipe's directory`() {
        test {
            every { fileService.listDirectoryFiles(any()) } returns recipeImagesFiles

            val foundImagesNames = recipeImageService.getAllImages(recipe)

            assertEquals(recipeImagesNames, foundImagesNames)
        }
    }

    @Test
    fun `getAllImages() returns an empty Set when the recipe's directory does not exists`() {
        test {
            every { fileService.listDirectoryFiles(any()) } returns emptySet()

            assertTrue {
                recipeImageService.getAllImages(recipe).isEmpty()
            }
        }
    }

    // download()

    @Test
    fun `download() writes the given image to the FileService and returns its name`() {
        test {
            val mockImage = MockMultipartFile("image.jpg", byteArrayOf(*"Random data".encodeToByteArray()))
            val expectedImageId = recipeImagesIds.maxOrNull()!! + 1L
            val expectedImageName = expectedImageId.imageName
            val expectedImagePath = expectedImageName.imagePath

            every { fileService.listDirectoryFiles(any()) } returns recipeImagesFiles
            every { fileService.writeToDirectory(any(), any(), any(), any()) } just runs

            val foundImageName = recipeImageService.download(mockImage, recipe)

            assertEquals(expectedImageName, foundImageName)

            verify {
                fileService.writeToDirectory(mockImage, expectedImagePath, any(), true)
            }
        }
    }

    // delete()

    @Test
    fun `delete() deletes the image with the given name in the FileService`() {
        test {
            val imageName = recipeImagesIds.first().imageName
            val imagePath = imageName.imagePath

            every { fileService.deleteFromDirectory(any(), any()) } just runs

            recipeImageService.delete(recipe, imageName)

            verify {
                fileService.deleteFromDirectory(imagePath, any())
            }
        }
    }
}
